#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Test the fprint function."""

import colorise
from io import StringIO
import pytest
import os
import sys


def test_valid_fprint():
    colorise.fprint('Hello {fg=red}world')
    colorise.fprint('Hello {fg=201}world')
    colorise.fprint('Hello {fg=#a696ff}world')
    colorise.fprint('Hello {fg=0xa696ff}world')
    colorise.fprint('Hello {fg=hls(0.6923;0.7960;1.0)}world')
    colorise.fprint('Hello {fg=hsv(249;41;100)}world')
    colorise.fprint('Hello {fg=rgb(167;151;255)}world')

    colorise.fprint('Hello {bg=red}world')
    colorise.fprint('Hello {bg=201}world')
    colorise.fprint('Hello {bg=#a696ff}world')
    colorise.fprint('Hello {bg=0xa696ff}world')
    colorise.fprint('Hello {bg=hls(0.6923;0.7960;1.0)}world')
    colorise.fprint('Hello {bg=hsv(249;41;100)}world')
    colorise.fprint('Hello {bg=rgb(167;151;255)}world')

    colorise.fprint('Hello {fg=red ,bg=red}world')
    colorise.fprint('Hello { fg=201,bg=201}world')
    colorise.fprint('Hello {fg=0xa696ff,bg=0xa696ff}world')
    colorise.fprint('Hello {fg=hls(0.6923;0.7960;1.0),'
                    'bg=hls(0.6923;0.7960;1.0)}world')
    colorise.fprint('Hello {fg=hsv(249;41;100),bg=hsv(249;41;100)}world')
    colorise.fprint('Hello {fg=rgb(167;151;255),bg=rgb(167;151;255)}world')


def test_invalid_fprint():
    kwargs = [
        'unknown',
        '300',
        '#a69ff',
        '0xa69ff',
        'hls(0.6923,0.7960;1.0=',
        'hsv(249;41;-100)',
        'rgb(167;xxx;255)',
    ]

    for kwarg in kwargs:
        with pytest.raises(ValueError):
            colorise.fprint('{fg=' + kwarg + '}Hello')

        with pytest.raises(ValueError):
            colorise.fprint('{bg=' + kwarg + '}Hello')


def test_duplicate_color_spec():
    with pytest.raises(ValueError, match='foreground'):
        colorise.fprint('Hello {fg=red,fg=red}world')

    with pytest.raises(ValueError, match='background'):
        colorise.fprint('Hello {bg=red,bg=red}world')


@pytest.mark.skip_on_windows
def test_valid_named_fprint_output():
    sio = StringIO()

    with pytest.redirect_stdout(sio):
        colorise.fprint('{fg=red}Hello', file=sys.stdout)
        assert sio.getvalue() == '\x1b[0m\x1b[31mHello\x1b[0m' + os.linesep


@pytest.mark.require_colors(256)
def test_valid_256_index_fprint_output():
    sio = StringIO()

    with pytest.redirect_stdout(sio):
        colorise.fprint('{fg=201}Hello', file=sys.stdout)
        assert sio.getvalue() == '\x1b[0m\x1b[38;5;201mHello\x1b[0m'\
            + os.linesep


@pytest.mark.require_colors(256**3)
def test_valid_truecolor_fprint_output():
    tests = [
        ('fg=0xa696ff',
         '\x1b[0m\x1b[38;2;166;150;255mHello\x1b[0m' + os.linesep),
        ('fg=0xa696ff',
         '\x1b[0m\x1b[38;2;166;150;255mHello\x1b[0m' + os.linesep),
        ('fg=hsv(249;41;100)',
         '\x1b[0m\x1b[38;2;166;150;255mHello\x1b[0m' + os.linesep),
        ('fg=rgb(167;151;255)',
         '\x1b[0m\x1b[38;2;167;151;255mHello\x1b[0m' + os.linesep),
    ]

    for color, expected_result in tests:
        sio = StringIO()

        with pytest.redirect_stdout(sio):
            colorise.fprint('{' + color + '}Hello', file=sys.stdout)
            assert sio.getvalue() == expected_result


@pytest.mark.skip_on_windows
def test_fprint_disabled():
    sio = StringIO()

    with pytest.redirect_stdout(sio):
        colorise.fprint('{fg=red}Hello', file=sys.stdout, enabled=False)
        assert sio.getvalue() == '\x1b[0mHello' + os.linesep


@pytest.mark.skip_on_windows
def test_fprint_proper_reset():
    sio = StringIO()

    with pytest.redirect_stdout(sio):
        colorise.set_color(fg='red')
        colorise.fprint('Hel{bg=blue}lo', file=sys.stdout)
        assert sio.getvalue() == '\x1b[0mHel\x1b[44mlo\x1b[0m' + os.linesep


@pytest.mark.skip_on_windows
def test_fprint_autoreset():
    text = '{fg=red}Hello {bg=blue}world!'
    sio = StringIO()

    with pytest.redirect_stdout(sio):
        colorise.fprint(text, file=sys.stdout, autoreset=False)
        assert sio.getvalue() ==\
            '\x1b[0m\x1b[31mHello \x1b[44mworld!\x1b[0m' + os.linesep

    sio = StringIO()

    with pytest.redirect_stdout(sio):
        colorise.fprint(text, file=sys.stdout, autoreset=True)
        assert sio.getvalue() ==\
            '\x1b[0m\x1b[31mHello \x1b[0m\x1b[44mworld!\x1b[0m' + os.linesep



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Windows color functions."""

from colorise.attributes import Attr
from colorise.cluts import get_color
import colorise.nix.color_functions
from colorise.win.cluts import can_interpret_ansi
from colorise.win.win32_functions import\
    get_win_handle,\
    set_console_text_attribute,\
    redefine_colors as _redefine_colors
import functools
import operator
import os
import platform
import sys


def num_colors():
    """Get the number of colors supported by the terminal."""
    if os.environ.get('ConEmuANSI', '') == 'ON':
        # ANSI escapes sequences are interpreted. ConEmu console detected which
        # supports 24-bit colors, but can we detect this somehow?
        #
        # NOTE: ConEmu also supports more attributes than the normal Windows
        # console.
        return 256

    release = platform.win32_ver()[0]
    build = sys.getwindowsversion()[2]

    # Windows 10 build 14931 has support for 24-bit colors
    if release == '10' and build >= 14931:
        return 2**24

    if can_interpret_ansi():
        return 2**24

    # Supported colors in Windows are pre-determined. Though you can update the
    # colors in the color table on Vista and beyond, this also changes all
    # colors of text already in the console window
    return 16


if num_colors() > 16 and can_interpret_ansi():
    # Extended terminal capabilities for interpreting ANSI escape codes,
    # delegate to nix color functions

    def reset_color(file=sys.stdout):
        """Reset all colors and attributes."""
        colorise.nix.color_functions.reset_color(file)

    def set_color(fg=None, bg=None, attributes=[], file=sys.stdout):
        """Set color and attributes in the terminal."""
        colorise.nix.color_functions.set_color(fg, bg, attributes, file,
                                               num_colors_func=num_colors)
else:
    # Ordinary terminal capabilities, use Windows API

    def reset_color(file=sys.stdout):
        """Reset all colors and attributes."""
        handle = get_win_handle(file)
        set_console_text_attribute(handle,
                                   handle.default_fg | handle.default_bg)

    def or_bit_flags(*bit_flags):
        """Bitwise OR together a list of bitflags into a single flag."""
        return functools.reduce(operator.or_, bit_flags)

    def set_color(fg=None, bg=None, attributes=[], file=sys.stdout):
        """Set color and attributes in the terminal."""
        if fg or bg or attributes:
            if Attr.Reset not in attributes:
                handle = get_win_handle(file)
                color_count = num_colors()
                codes = []

                codes.extend(get_color(fg, color_count, colorise.win.cluts,
                                       False, attributes)
                             if fg else [handle.default_fg])
                codes.extend(get_color(bg, color_count, colorise.win.cluts,
                                       True, attributes)
                             if bg else [handle.default_bg])

                # Combine attributes and color codes into a single bitflag
                flags = or_bit_flags(*codes)
                set_console_text_attribute(handle, flags)
            else:
                reset_color(file)


def redefine_colors(color_map, file=sys.stdout):
    """Redefine the base console colors with a new mapping."""
    _redefine_colors(color_map, file)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

sys.path.insert(0, os.path.abspath('..'))

# -- Project information -----------------------------------------------------

project = 'colorise'
copyright = '2019, Alexander Asp Bock'
author = 'Alexander Asp Bock'
master_doc = 'index'

# The full version, including alpha/beta/rc tags
release = '1.0.1'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.linkcode',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    '_build',
    'Thumbs.db',
    '.DS_Store',
]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

html_theme_options = {
    'description':     'A Python module for printing colored text in terminals.',
    'logo':            '../_images/colorise-logo.png',
    'github_user':     'MisanthropicBit',
    'github_repo':     'colorise',
    'github_banner':   'true',
    'github_button':   'false',
    'extra_nav_links': {
        'Issues':        'https://github.com/MisanthropicBit/colorise/issues',
        'Pull Requests': 'https://github.com/MisanthropicBit/colorise/pulls',
    },
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

base_url = 'https://github.com/{github_user}/{github_repo}/blob/{branch}'\
    .format(branch='master', **html_theme_options)


def linkcode_resolve(domain, info):
    """Resolve links to github source code."""
    if domain != 'py' or '.' in info['fullname']:
        return None

    module_path = info['module'].replace('.', os.sep)

    if module_path == project:
        module_path = os.path.join(module_path, '__init__')

    return os.path.join(base_url, module_path + '.py')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Pytest configuration file."""

import colorise
import platform
import pytest
from contextlib import redirect_stdout

def pytest_configure(config):  # noqa
    pytest.redirect_stdout = redirect_stdout

    config.addinivalue_line(
        'markers',
        'skip_on_windows: skip test on windows'
    )

    config.addinivalue_line(
        'markers',
        'skip_on_nix: skip test on nix systems'
    )

    config.addinivalue_line(
        'markers',
        'require_colors(count): run test only if a specific number of colors '
        'are supported'
    )


_PLATFORM = platform.system().lower()


def pytest_runtest_setup(item):  # noqa
    for mark in item.iter_markers():
        if mark.name == 'skip_on_windows':
            if _PLATFORM.startswith('win'):
                pytest.skip('Test skipped on Windows systems')
        elif mark.name == 'skip_on_nix':
            if not _PLATFORM.startswith('win'):
                pytest.skip('Test skipped on nix systems')
        elif mark.name == 'require_colors':
            required_num_colors = mark.args[0]

            if colorise.num_colors() < mark.args[0]:
                msg = 'Test only relevant for terminals with {0}'

                if required_num_colors == 256**3:
                    color_name = 'true-color'
                else:
                    color_name = str(required_num_colors) + ' colors'

                pytest.skip(msg.format(color_name))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Linux/Mac color functions."""

import os
import sys
import colorise.error
import colorise.nix.cluts
from colorise.attributes import Attr
from colorise.cluts import get_color
from colorise.terminal import terminal_name


def num_colors():
    """Attempt to get the number of colors supported by the terminal."""
    # iTerm supports true-color from version 3 onward, earlier versions
    # supported 256 colors
    if terminal_name() == 'iTerm.app':
        version = os.environ.get('TERM_PROGRAM_VERSION', '')

        if version and int(version.split('.')[0]) > 2:
            return 2**24

        return 256

    # If all else fails, use curses
    import curses
    color_count = 0

    try:
        # Use the file descriptor of the original value of sys.stdout in case
        # it has been redirected by pytest, tox or something else
        curses.setupterm(fd=sys.__stdout__.fileno())
        color_count = curses.tigetnum('colors')
    except curses.error:
        pass

    if color_count <= 0:
        # Failed to get color count from curses, default to 16 colors
        return 16

    return color_count


def to_ansi(*codes):
    """Convert a set of ANSI codes into a valid ANSI escape sequence."""
    if not codes:
        return ''

    return colorise.nix.cluts._COLOR_ESCAPE_CODE +\
        '{0}m'.format(';'.join(str(c) for c in codes))


def attributes_to_codes(attributes):
    """Convert a set of attributes to ANSI escape codes."""
    return [int(attr.value) for attr in attributes]


def reset_color(file=sys.stdout):
    """Reset all colors and attributes."""
    file.write(to_ansi(Attr.Reset.value))


def set_color(fg=None, bg=None, attributes=[], file=sys.stdout,
              num_colors_func=num_colors):
    """Set color and attributes of the terminal.

    'fg' and 'bg' specify foreground- and background colors while 'attributes'
    is a list of desired attributes. The 'file' keyword specifies the target
    output stream.

    """
    codes = []

    if attributes:
        codes.append(to_ansi(*attributes_to_codes(attributes)))

    if Attr.Reset not in attributes:
        color_count = num_colors_func()

        for colorspec, isbg in ((fg, False), (bg, True)):
            if colorspec:
                prefix, color = get_color(colorspec, color_count,
                                          colorise.nix.cluts, isbg)
                codes.append(prefix.format(color))

    if codes:
        file.write(''.join(codes))


def redefine_colors(color_map, file=sys.stdout):
    """Redefine the base console colors with a new mapping."""
    raise colorise.error.NotSupportedError('Cannot redefine colors on nix '
                                           'systems')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Demonstration of colorise's autoreset future when exceptions are thrown."""

import colorise


if __name__ == '__main__':
    # Set the current color
    colorise.set_color('blue')

    # Should be printed in blue
    print("Chapter 1: The Zen of Python")

    try:
        # Even though this statement will fail, colors will be reset
        colorise.fprint("{fg=???}I can wait to be in color!")
    except ValueError:
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Windows color look-up tables and functions."""

from colorise.attributes import Attr
from colorise.color_tools import closest_color
from colorise.win.win32_functions import\
    can_interpret_ansi,\
    can_redefine_colors,\
    get_windows_clut
import colorise.nix.cluts

# Character attributes as defined in wincon.h
# We cannot get these through ctypes since they are #defines
_WIN_ATTRIBUTES = {
        Attr.Reset:     0x0000,
        Attr.Bold:      0x0008,
        Attr.Intense:   0x0008,  # Alias
        Attr.Faint:     0x0000,
        Attr.Italic:    0x0000,
        Attr.Underline: 0x8000,
        Attr.Blink:     0x0000,
        Attr.Reverse:   0x4000,
    }

if can_redefine_colors():
    _WINDOWS_CLUT = get_windows_clut()
else:
    # Windows 16-color (logical) look-up table
    _WINDOWS_CLUT = {
        0:  (0x00, 0x00, 0x00),  # Black
        1:  (0x00, 0x00, 0x80),  # Red
        2:  (0x00, 0x80, 0x00),  # Green
        3:  (0x00, 0x80, 0x80),  # Yellow
        4:  (0x80, 0x00, 0x00),  # Magenta
        5:  (0x80, 0x00, 0x80),  # Blue
        6:  (0x80, 0x80, 0x00),  # Cyan
        7:  (0xff, 0xff, 0xff),  # White
        8:  (0x80, 0x80, 0x80),  # Gray/intensity
        # The remaining colors are sometimes referred to as the 'light' colors
        9:  (0x00, 0x00, 0xff),  # Light blue
        10: (0x00, 0xff, 0x00),  # Light green
        11: (0x00, 0xff, 0xff),  # Light cyan
        12: (0xff, 0x00, 0x00),  # Light red
        13: (0xff, 0x00, 0xff),  # Light purple
        14: (0xff, 0xff, 0x00),  # Light yellow
        15: (0xff, 0xff, 0xff)   # Light white
    }

_FOREGROUND_RED = 0x0004
_FOREGROUND_GREEN = 0x0002
_FOREGROUND_BLUE = 0x0001
_FOREGROUND_INTENSITY = _WIN_ATTRIBUTES[Attr.Bold]

# List of logical colors names on Windows (as defined by colorise)
_WINDOWS_LOGICAL_NAMES = {
    'black':   0,
    'red':     _FOREGROUND_RED,
    'green':   _FOREGROUND_GREEN,
    'yellow':  _FOREGROUND_RED | _FOREGROUND_GREEN,
    'magenta': _FOREGROUND_RED | _FOREGROUND_BLUE,
    'purple':  _FOREGROUND_RED | _FOREGROUND_BLUE,
    'blue':    _FOREGROUND_BLUE,
    'cyan':    _FOREGROUND_GREEN | _FOREGROUND_BLUE,
    'white':   _FOREGROUND_RED | _FOREGROUND_GREEN | _FOREGROUND_BLUE |
               _FOREGROUND_INTENSITY,
}

for name in ['red', 'green', 'blue', 'yellow', 'purple', 'magenta', 'cyan']:
    _WINDOWS_LOGICAL_NAMES['light' + name] =\
        _WINDOWS_LOGICAL_NAMES[name] | _FOREGROUND_INTENSITY

_WINDOWS_LOGICAL_NAMES['gray'] = _WINDOWS_LOGICAL_NAMES['black'] |\
    _FOREGROUND_INTENSITY
_WINDOWS_LOGICAL_NAMES['grey'] = _WINDOWS_LOGICAL_NAMES['black'] |\
    _FOREGROUND_INTENSITY
_WINDOWS_LOGICAL_NAMES['lightgrey'] = _FOREGROUND_RED | _FOREGROUND_GREEN |\
    _FOREGROUND_BLUE
_WINDOWS_LOGICAL_NAMES['lightgray'] = _WINDOWS_LOGICAL_NAMES['lightgrey']


def get_clut(color_count):
    """Return the appropriate color look-up table."""
    return _WINDOWS_CLUT


def to_codes(bg, color, attributes):
    """Convert a set of attributes to Windows character attributes."""
    codes = [color] + [_WIN_ATTRIBUTES[attr] for attr in attributes]

    if bg:
        return [code << 4 for code in codes]

    return codes


def color_from_name(name, color_count, bg, attributes):
    """Return the color value and color count for a given color name."""
    if name not in _WINDOWS_LOGICAL_NAMES:
        raise ValueError("Unknown color name '{0}'".format(name))

    return to_codes(bg, _WINDOWS_LOGICAL_NAMES[name], attributes)


def color_from_index(idx, color_count, bg, attributes):
    """Return the color value and color count for a given color index."""
    if can_interpret_ansi():
        # We can interpret ANSI escape sequences, delegate to nix function
        return colorise.nix.cluts.color_from_index(idx, color_count, bg,
                                                   attributes)

    if idx < 0 or idx > 255:
        raise ValueError('Color index must be in range 0-255 inclusive')

    if idx < 16:
        color = closest_color(
            colorise.nix.cluts._NIX_SYSTEM_COLORS,
            _WINDOWS_CLUT
        )
    elif idx < 88:
        # 88 color index
        color = closest_color(
            colorise.nix.cluts._XTERM_CLUT_88[idx],
            _WINDOWS_CLUT
        )
    elif idx < 256:
        # 256 color index
        color = closest_color(
            colorise.nix.cluts._XTERM_CLUT_256[idx],
            _WINDOWS_CLUT
        )

    return to_codes(bg, color, attributes)


def get_rgb_color(color_count, bg, rgb, attributes):
    """Get the color for an RGB triple or approximate it if necessary."""
    if color_count == 2**24:
        # We have true-color capabilities, delegate to nix function
        # approximate to closest color given current capabilities
        return colorise.nix.cluts.get_rgb_color(color_count, bg, rgb,
                                                attributes)

    # No true-color capabilities, approximate the rgb color
    idx = closest_color(rgb, get_clut(color_count))

    return to_codes(bg, idx, attributes)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Windows API functions."""

import ctypes
from ctypes import windll, wintypes, WinError
import os
import sys
from colorise.win.winhandle import WinHandle


# Invalid handle type for error checking
INVALID_HANDLE_VALUE = wintypes.HANDLE(-1).value

# Handle IDs for stdout and stderr
_STDOUT_HANDLE_ID = -11
_STDERR_HANDLE_ID = -12

# Console modes for console virtual terminal sequences
DISABLE_NEWLINE_AUTO_RETURN = 0x0008
ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x0004


# Struct defined in wincon.h
class CONSOLE_SCREEN_BUFFER_INFO(ctypes.Structure):
    _fields_ = [('dwSize',              wintypes._COORD),
                ('dwCursorPosition',    wintypes._COORD),
                ('wAttributes',         ctypes.c_ushort),
                ('srWindow',            wintypes._SMALL_RECT),
                ('dwMaximumWindowSize', wintypes._COORD)]


# Struct defined in wincon.h
class CONSOLE_SCREEN_BUFFER_INFOEX(ctypes.Structure):
    _fields_ = [('cbSize',               wintypes.ULONG),
                ('dwSize',               wintypes._COORD),
                ('dwCursorPosition',     wintypes._COORD),
                ('wAttributes',          ctypes.c_ushort),
                ('srWindow',             wintypes._SMALL_RECT),
                ('dwMaximumWindowSize',  wintypes._COORD),
                ('wPopupAttributes',     wintypes.WORD),
                ('bFullscreenSupported', wintypes.BOOL),
                ('ColorTable',           wintypes.COLORREF * 16)]


if not hasattr(wintypes, 'LPDWORD'):
    LPDWORD = ctypes.POINTER(wintypes.DWORD)
else:
    LPDWORD = wintypes.LPDWORD

# Set argument and return types for Windows API calls
windll.kernel32.GetConsoleScreenBufferInfo.argtypes =\
    [wintypes.HANDLE, ctypes.POINTER(CONSOLE_SCREEN_BUFFER_INFO)]
windll.kernel32.GetConsoleScreenBufferInfo.restype = wintypes.BOOL
windll.kernel32.GetStdHandle.argtypes = [wintypes.DWORD]
windll.kernel32.GetStdHandle.restype = wintypes.HANDLE
windll.kernel32.GetConsoleMode.argtypes = [wintypes.HANDLE, LPDWORD]
windll.kernel32.GetConsoleMode.restype = wintypes.BOOL
windll.kernel32.SetConsoleMode.argtypes = [wintypes.HANDLE, wintypes.DWORD]
windll.kernel32.SetConsoleMode.restype = wintypes.BOOL
windll.kernel32.GetLastError.argtypes = []
windll.kernel32.GetLastError.restype = wintypes.DWORD
windll.kernel32.SetLastError.argtypes = [wintypes.DWORD]
windll.kernel32.SetLastError.restype = None  # void
windll.kernel32.FormatMessageW.argtypes = [wintypes.DWORD,
                                           wintypes.LPCVOID,
                                           wintypes.DWORD,
                                           wintypes.DWORD,
                                           wintypes.LPWSTR,
                                           wintypes.DWORD,
                                           wintypes.LPVOID]
windll.kernel32.FormatMessageW.restype = wintypes.DWORD
windll.kernel32.LocalFree.argtypes = [wintypes.HLOCAL]
windll.kernel32.LocalFree.restype = wintypes.HLOCAL
windll.kernel32.SetConsoleTextAttribute.argtypes = [wintypes.HANDLE,
                                                    wintypes.WORD]
windll.kernel32.SetConsoleTextAttribute.restype = wintypes.BOOL


def can_redefine_colors():
    """Return whether the terminal allows redefinition of colors."""
    return windll.kernel32.SetConsoleScreenBufferInfoEx is not None


if can_redefine_colors():
    # We can query RGB values of console colors on Windows
    windll.kernel32.GetConsoleScreenBufferInfoEx.argtypes =\
        [wintypes.HANDLE, ctypes.POINTER(CONSOLE_SCREEN_BUFFER_INFOEX)]
    windll.kernel32.GetConsoleScreenBufferInfoEx.restype = wintypes.BOOL


def create_std_handle(handle_id):
    """Create a Windows standard handle from an identifier."""
    win_handle = WinHandle(windll.kernel32.GetStdHandle(handle_id))

    if win_handle.handle == INVALID_HANDLE_VALUE:
        raise WinError()

    csbi = CONSOLE_SCREEN_BUFFER_INFO()

    if windll.kernel32.GetConsoleScreenBufferInfo(
                win_handle.handle,
                ctypes.byref(csbi)) == 0:
        raise WinError()

    # Set defaults color values
    # TODO: Do these need to be reread when colors are redefined?
    win_handle.default_fg = csbi.wAttributes & 0xf
    win_handle.default_bg = (csbi.wAttributes >> 4) & 0xf

    # Set the color for the handle
    win_handle.fg = win_handle.default_fg
    win_handle.bg = win_handle.default_bg

    return win_handle


# Create handles for stdout and stderr
_STDOUT_HANDLE = create_std_handle(-11)
_STDERR_HANDLE = create_std_handle(-12)


def get_win_handle(target):
    """Return the Windows handle corresponding to a Python handle."""
    if target is sys.stdout:
        return _STDOUT_HANDLE
    elif target is sys.stderr:
        return _STDERR_HANDLE

    raise ValueError('Only stdout and stderr supported')


def get_windows_clut():
    """Query and return the internal Windows color look-up table."""
    # On Windows Vista and beyond you can query the current colors in the
    # color table. On older platforms, use the default color table
    csbiex = CONSOLE_SCREEN_BUFFER_INFOEX()
    csbiex.cbSize = ctypes.sizeof(CONSOLE_SCREEN_BUFFER_INFOEX)

    if windll.kernel32.GetConsoleScreenBufferInfoEx(
                get_win_handle(sys.stdout).handle,
                ctypes.byref(csbiex)
            ) == 0:
        raise WinError()

    clut = {}

    # Update according to the currently set colors
    for i in range(16):
        clut[i] =\
            (csbiex.ColorTable[i] & 0xff,
             (csbiex.ColorTable[i] >> 8) & 0xff,
             (csbiex.ColorTable[i] >> 16) & 0xff)

    return clut


def enable_virtual_terminal_processing(handle):
    """Enable Windows processing of ANSI escape sequences."""
    if not handle or handle == INVALID_HANDLE_VALUE:
        raise ValueError('Invalid handle')

    console_mode = wintypes.DWORD(0)

    if windll.kernel32.GetConsoleMode(handle.handle,
                                      ctypes.byref(console_mode)) == 0:
        raise WinError()

    handle.console_mode = console_mode

    target_mode = wintypes.DWORD(console_mode.value |
                                 ENABLE_VIRTUAL_TERMINAL_PROCESSING |
                                 DISABLE_NEWLINE_AUTO_RETURN)

    # First attempt to set console mode to interpret ANSI escape codes and
    # disable immediately jumping to the next console line
    if windll.kernel32.SetConsoleMode(handle.handle, target_mode) == 0:
        # If that fails, try just setting the mode for ANSI escape codes
        target_mode = wintypes.DWORD(console_mode.value |
                                     ENABLE_VIRTUAL_TERMINAL_PROCESSING)

        if windll.kernel32.SetConsoleMode(handle.handle, target_mode) == 0:
            return None

    # Return the original console mode so we can restore it later
    return console_mode


def restore_console_mode(handle, restore_mode):
    """Restore the console mode for a handle to its original mode."""
    if not handle or handle == INVALID_HANDLE_VALUE:
        raise ValueError('Invalid handle')

    if not windll.kernel32.SetConsoleMode(handle.handle, restore_mode):
        raise WinError()


def restore_console_modes():
    """Restore console modes for stdout and stderr to their original mode."""
    if can_interpret_ansi():
        stdout = get_win_handle(sys.stdout)
        stderr = get_win_handle(sys.stderr)
        restore_console_mode(stdout, stdout.console_mode)
        restore_console_mode(stderr, stderr.console_mode)


_WIN_CAN_INTERPRET_ANSI_CODES =\
    enable_virtual_terminal_processing(_STDOUT_HANDLE) is not None

enable_virtual_terminal_processing(_STDERR_HANDLE)


def can_interpret_ansi():
    """Return True if the Windows console can interpret ANSI escape codes."""
    if os.environ.get('ConEmuANSI', '') == 'ON':
        return True

    return _WIN_CAN_INTERPRET_ANSI_CODES


def set_console_text_attribute(handle, flags):
    """Set the console's text attributes."""
    if not handle or handle == INVALID_HANDLE_VALUE:
        raise ValueError('Invalid handle')

    if windll.kernel32.SetConsoleTextAttribute(handle.handle,
                                               wintypes.WORD(flags)) == 0:
        raise WinError()


def encode_rgb_tuple(rgb):
    """Hexadecimally encode an rgb tuple as 0xbbggrr."""
    r, g, b = rgb

    return (b << 16) | (g << 8) | r


def redefine_colors(color_map, file=sys.stdout):
    """Redefine the base console colors with a new mapping.

    This only redefines the 8 colors in the console and changes all text in the
    console that already uses the logical names. E.g. if 'red' is mapped to the
    color red and this function changes it to another color, all text in 'red'
    will be rendered with this new color, even though it may already have been
    written to the console.

    """
    if not can_redefine_colors():
        raise RuntimeError('Cannot redefine colors on this system')

    if not all(0 <= c < 16 for c in color_map):
        raise RuntimeError('New color map must contain indices in range 0-15')

    # Create a new CONSOLE_SCREEN_BUFFER_INFOEX structure based on the given
    # color map
    csbiex = CONSOLE_SCREEN_BUFFER_INFOEX()

    # We must set the size of the structure before using it
    csbiex.cbSize = ctypes.sizeof(CONSOLE_SCREEN_BUFFER_INFOEX)

    win_handle = get_win_handle(file)

    # Get console color info
    if windll.kernel32.GetConsoleScreenBufferInfoEx(
                win_handle.handle,
                ctypes.byref(csbiex)
            ) == 0:
        raise WinError()

    # Redefine colortable
    for idx in color_map:
        csbiex.ColorTable[idx] = encode_rgb_tuple(color_map[idx])

    # Set the new colors
    if windll.kernel32.SetConsoleScreenBufferInfoEx(
            win_handle.handle, csbiex) == 0:
        raise WinError()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Test attributes."""

import colorise
from colorise.attributes import Attr
from io import StringIO
import pytest
import os
import sys


@pytest.fixture
def expected_results():
    return zip(Attr, [
            '\x1b[0m\x1b[0mHello\x1b[0m' + os.linesep,
            '\x1b[0m\x1b[1m\x1b[31mHello\x1b[0m' + os.linesep,
            '\x1b[0m\x1b[2m\x1b[31mHello\x1b[0m' + os.linesep,
            '\x1b[0m\x1b[3m\x1b[31mHello\x1b[0m' + os.linesep,
            '\x1b[0m\x1b[4m\x1b[31mHello\x1b[0m' + os.linesep,
            '\x1b[0m\x1b[5m\x1b[31mHello\x1b[0m' + os.linesep,
            '\x1b[0m\x1b[7m\x1b[31mHello\x1b[0m' + os.linesep,
        ])


def test_attributes():
    colorise.cprint('Hello',                       attributes=[Attr.Bold])
    colorise.cprint('Hello', fg='red',             attributes=[Attr.Bold])
    colorise.cprint('Hello',            bg='blue', attributes=[Attr.Bold])
    colorise.cprint('Hello', fg='red',  bg='blue', attributes=[Attr.Bold])

    colorise.fprint('{bold}Hello')
    colorise.fprint('{fg=red,bold}Hello')
    colorise.fprint('{bg=blue,bold}Hello')
    colorise.fprint('{fg=red,bg=blue,bold}Hello')


def test_attribute_aliases():
    colorise.cprint('Hello',                       attributes=[Attr.Intense])
    colorise.cprint('Hello', fg='red',             attributes=[Attr.Intense])
    colorise.cprint('Hello',            bg='blue', attributes=[Attr.Intense])
    colorise.cprint('Hello', fg='red',  bg='blue', attributes=[Attr.Intense])

    colorise.fprint('{intense}Hello')
    colorise.fprint('{fg=red,intense}Hello')
    colorise.fprint('{bg=blue,intense}Hello')
    colorise.fprint('{fg=red,bg=blue,intense}Hello')


def test_invalid_attributes():
    with pytest.raises(ValueError):
        colorise.fprint('{overlined}Hello')

    with pytest.raises(ValueError):
        colorise.fprint('{fg=red,overlined}Hello')

    with pytest.raises(ValueError):
        colorise.fprint('{bg=blue,overlined}Hello')

    with pytest.raises(ValueError):
        colorise.fprint('{fg=red,bg=blue,overlined}Hello')


@pytest.mark.skip_on_windows
def test_attribute_cprint_output(expected_results):
    for attribute, result in expected_results:
        sio = StringIO()

        with pytest.redirect_stdout(sio):
            colorise.cprint('Hello', fg='red', attributes=[attribute],
                            file=sys.stdout)
            assert sio.getvalue() == result


@pytest.mark.skip_on_windows
def test_attribute_fprint_output(expected_results):
    for attribute, result in expected_results:
        sio = StringIO()

        with pytest.redirect_stdout(sio):
            colorise.fprint('{{fg=red,{0}}}Hello'
                            .format(attribute.name.lower()),
                            file=sys.stdout)
            assert sio.getvalue() == result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Test the cprint function."""

import colorise
from io import StringIO
import pytest
import os
import sys


def test_valid_cprint():
    colorise.cprint('Hello', fg='red')
    colorise.cprint('Hello', fg=201)
    colorise.cprint('Hello', fg='#a696ff')
    colorise.cprint('Hello', fg='0xa696ff')
    colorise.cprint('Hello', fg='hls(0.6919;0.7940;1.0)')
    colorise.cprint('Hello', fg='hsv(249;41;100)')
    colorise.cprint('Hello', fg='rgb(167;151;255)')

    colorise.cprint('Hello', bg='red')
    colorise.cprint('Hello', bg=201)
    colorise.cprint('Hello', bg='#a696ff')
    colorise.cprint('Hello', bg='0xa696ff')
    colorise.cprint('Hello', bg='hls(0.6919;0.7940;1.0)')
    colorise.cprint('Hello', bg='hsv(249;41;100)')
    colorise.cprint('Hello', bg='rgb(167;151;255)')

    colorise.cprint('Hello', fg='red', bg='red')
    colorise.cprint('Hello', fg=201, bg=201)
    colorise.cprint('Hello', fg='#a696ff', bg='#a696ff')
    colorise.cprint('Hello', fg='0xa696ff', bg='0xa696ff')
    colorise.cprint('Hello', fg='hls(0.6919;0.7940;1.0)',
                    bg='hls(0.6919;0.7940;1.0)')
    colorise.cprint('Hello', fg='hsv(249;41;100)', bg='hsv(249;41;100)')
    colorise.cprint('Hello', fg='rgb(167;151;255)', bg='rgb(167;151;255)')


def test_invalid_cprint():
    kwargs = [
        {'fg': 'unknown'},
        {'fg': 256},
        {'bg': '#a69ff'},
        {'bg': '0xa69ff'},
        {'fg': 'hls(0.6919,0.7940;1.0='},
        {'bg': 'hsv(249;41;-100)'},
        {'fg': 'rgb(167;xxx;255)'},
    ]

    for kwarg in kwargs:
        with pytest.raises(ValueError):
            colorise.cprint('Hello', **kwarg)


@pytest.mark.skip_on_windows
def test_valid_named_cprint_output():
    tests = zip(
        ['red', 'red', None],
        [None, 'blue', 'blue'],
        [
            '\x1b[0m\x1b[31mHello\x1b[0m' + os.linesep,
            '\x1b[0m\x1b[31m\x1b[44mHello\x1b[0m' + os.linesep,
            '\x1b[0m\x1b[44mHello\x1b[0m' + os.linesep
        ]
    )

    for fg, bg, result in tests:
        sio = StringIO()

        with pytest.redirect_stdout(sio):
            colorise.cprint('Hello', fg=fg, bg=bg, file=sys.stdout)
            assert sio.getvalue() == result


@pytest.mark.require_colors(256)
def test_valid_256_index_cprint_output():
    sio = StringIO()

    with pytest.redirect_stdout(sio):
        colorise.cprint('Hello', fg=201, file=sys.stdout)
        assert sio.getvalue() == '\x1b[0m\x1b[38;5;201mHello\x1b[0m'\
            + os.linesep


@pytest.mark.require_colors(256**3)
def test_valid_truecolor_cprint_output():
    tests = [
        ({'fg': '0xa696ff'},
         '\x1b[0m\x1b[38;2;166;150;255mHello\x1b[0m' + os.linesep),
        ({'fg': 'hls(0.6919;0.7940;1.0)'},
         '\x1b[0m\x1b[38;2;166;150;255mHello\x1b[0m' + os.linesep),
        ({'fg': 'hsv(249;41;100)'},
         '\x1b[0m\x1b[38;2;166;150;255mHello\x1b[0m' + os.linesep),
        ({'fg': 'rgb(167;151;255)'},
         '\x1b[0m\x1b[38;2;167;151;255mHello\x1b[0m' + os.linesep),
    ]

    for kwargs, expected_result in tests:
        sio = StringIO()

        with pytest.redirect_stdout(sio):
            colorise.cprint('Hello', file=sys.stdout, **kwargs)
            assert sio.getvalue() == expected_result


@pytest.mark.skip_on_windows
def test_cprint_disabled():
    sio = StringIO()

    with pytest.redirect_stdout(sio):
        colorise.cprint('Hello', fg='red', file=sys.stdout, enabled=False)
        assert sio.getvalue() == '\x1b[0mHello' + os.linesep


@pytest.mark.skip_on_windows
def test_cprint_proper_reset():
    sio = StringIO()

    with pytest.redirect_stdout(sio):
        colorise.set_color(fg='red')
        colorise.cprint('Hello', bg='blue', file=sys.stdout)
        assert sio.getvalue() == '\x1b[0m\x1b[44mHello\x1b[0m' + os.linesep



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Python module for easy, cross-platform colored output to the terminal."""

import atexit
import itertools
import os
import platform
import sys
import colorise.formatter
from colorise.attributes import Attr

_SYSTEM_OS = platform.system().lower()

__author__ = 'Alexander Asp Bock'
__version__ = '1.0.1'
__license__ = 'BSD 3-Clause'
__all__ = [
    'can_redefine_colors',
    'redefine_colors',
    'color_names',
    'num_colors',
    'set_color',
    'reset_color',
    'cprint',
    'fprint',
    'highlight'
]

# Determine which platform-specific color manager to import
if _SYSTEM_OS.startswith('win'):
    from colorise.win.color_functions import\
        reset_color as _reset_color,\
        set_color as _set_color,\
        redefine_colors as _redefine_colors,\
        num_colors as _num_colors

    from colorise.win.win32_functions import\
        can_redefine_colors as _can_redefine_colors,\
        restore_console_modes

    # Ensure that the console mode set before colorise was loaded is restored
    atexit.register(restore_console_modes)
else:
    from colorise.nix.color_functions import\
        reset_color as _reset_color,\
        set_color as _set_color,\
        redefine_colors as _redefine_colors,\
        num_colors as _num_colors

    from colorise.nix.cluts import\
        can_redefine_colors as _can_redefine_colors


def num_colors():
    """Return the number of colors supported by the terminal."""
    return _num_colors()


def can_redefine_colors():
    """Return True if the terminal supports redefinition of colors.

    Only returns True for Windows 7/Vista and beyond as of now.

    """
    return _can_redefine_colors()


def redefine_colors(color_map, file=sys.stdout):
    """Redefine colors using a color map of indices and RGB tuples.

    .. note::

        It is not currently possible to redefine colors on Mac and Linux
        systems via colorise.

    """
    _redefine_colors(color_map, file)


def color_names():
    """Return a list of supported color names."""
    return [
        'black',
        'red',
        'green',
        'yellow',
        'blue',
        'purple',
        'magenta',
        'cyan',
        'gray',
        'grey',
        'lightgrey',
        'lightgray',
        'lightred',
        'lightgreen',
        'lightyellow',
        'lightblue',
        'lightpurple',
        'lightcyan',
        'white',
    ]


def set_color(fg=None, bg=None, attributes=[], file=sys.stdout):
    """Set the current colors.

    If no arguments are given, sets default colors.

    """
    _set_color(fg, bg, attributes, file)


def reset_color(file=sys.stdout):
    """Reset all colors and attributes."""
    _reset_color(file)


def cprint(string, fg=None, bg=None, attributes=[], end=os.linesep,
           file=sys.stdout, enabled=True):
    """Print a string to a target stream with colors and attributes.

    The fg and bg keywords specify foreground- and background colors while
    attributes is a list of desired attributes. The remaining two keyword
    arguments are the same as Python's built-in print function.

    Colors and attributes are reset before the function returns.

    """
    # Flush any remaining stuff before resetting colors
    file.flush()
    reset_color(file)

    if not enabled:
        file.write(string)
        file.write(end)
        file.flush()
    else:
        set_color(fg, bg, attributes, file)
        file.write(string)
        file.flush()  # Flush before resetting colors
        reset_color(file)

        # Make sure to print the end keyword after resetting so the next line
        # is not affected by a newline or similar
        file.write(end)
        file.flush()  # Flush before resetting colors


# Global color formatter instance
_COLOR_FORMATTER = colorise.formatter.ColorFormatter(set_color, reset_color)


def fprint(fmt, autoreset=True, end=os.linesep, file=sys.stdout, enabled=True):
    """Print a string with color formatting.

    The autoreset keyword controls if colors and attributes are reset before
    each new color format. For example:

    >>> colorise.fprint('{fg=blue}Hi {bg=red}world', autoreset=False)

    would print 'Hi' in blue foreground colors and 'world' in blue foreground
    colors AND a red background color, whereas:

    >>> colorise.fprint('{fg=blue}Hi {bg=red}world', autoreset=True)

    would print 'Hi' in blue foreground colors but 'world' only with a red
    background color since colors are reset when '{bg=red}' is encountered.

    The remaining two keyword arguments are the same as Python's built-in print
    function.

    Colors and attribtues are reset before the function returns.

    """
    _COLOR_FORMATTER.autoreset = autoreset
    _COLOR_FORMATTER.file = file
    _COLOR_FORMATTER.enabled = enabled
    _COLOR_FORMATTER.format(fmt)

    if enabled:
        file.flush()  # Flush before resetting colors
        reset_color(file)

    # Make sure to print the end keyword after resetting so the next line is
    # not affected by a newline or similar
    file.write(end)


def highlight(string, indices, fg=None, bg=None, attributes=[], end=os.linesep,
              file=sys.stdout, enabled=True):
    """Highlight characters using indices and print to a target stream.

    The indices argument is a list of indices (not necessarily sorted) for
    which to apply the colors and attributes. Indices that are out of bounds
    are ignored.

    fg and bg specify foreground- and background colors while attributes is a
    list of desired attributes. The remaining two keyword arguments are the
    same as Python's built-in print function.

    Colors and attribtues are reset before the function returns.

    """
    if not string or not indices or not (fg or bg or attributes)\
            or not enabled:
        file.write(string + end)
        return

    idx = 0

    # Group consecutive indices, e.g. [0, 2, 3, 5, 6] -> [(0), (2, 3), (5, 6)]
    # NOTE: The lambda syntax is necessary to support both Python 2 and 3
    groups = itertools.groupby(enumerate(sorted(indices)), lambda x: x[0]-x[1])

    # Flush any remaining stuff before resetting colors
    file.flush()
    reset_color(file)

    for _, group in groups:
        # Get the starting and ending indices of the group
        group = list(group)
        start_idx, end_idx = group[0][1], group[-1][1]+1

        # Write anything up until the start index of the current group
        file.write(string[idx:start_idx])
        file.flush()

        set_color(fg, bg, attributes, file)

        # Write the range of characters specified by the group
        file.write(string[start_idx:end_idx])
        file.flush()

        reset_color(file)

        # Set current index to end of group
        idx = end_idx

    # Write anything that is left to write
    if idx < len(string):
        file.write(string[idx:])

    file.write(end)


# Ensure colors and attributes return to normal when colorise is quit
if sys.stdout.isatty():
    atexit.register(reset_color)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Demonstrates how colorise automatically reset colors when quit."""

import colorise

IPSUM = """Lorem ipsum dolor sit amet, consectetuer adipiscing elit,
sed diam nonummy nibh euismod tincidunt ut laoreet dolore magna
aliquam erat volutpat. Ut wisi enim ad minim veniam, quis
nostrud exerci tation ullamcorper suscipit lobortis nisl ut
aliquip ex ea commodo consequat. Duis autem vel eum iriure dolor
in hendrerit in vulputate velit esse molestie consequat, vel
illum dolore eu feugiat nulla facilisis at vero eros et accumsan
et iusto odio dignissim qui blandit praesent luptatum zzril delenit
augue duis dolore te feugait nulla facilisi. Nam liber tempor
cum soluta nobis eleifend option congue nihil imperdiet doming
id quod mazim placerat facer possim assum. Typi non habent
claritatem insitam; est usus legentis in iis qui facit eorum
claritatem. Investigationes demonstraverunt lectores legere me
lius quod ii legunt saepius. Claritas est etiam processus
dynamicus, qui sequitur mutationem consuetudium lectorum.
Mirum est notare quam littera gothica, quam nunc putamus
parum claram, anteposuerit litterarum formas humanitatis per
seacula quarta decima et quinta decima. Eodem modo typi, qui
nunc nobis videntur parum clari, fiant sollemnes in futurum."""

if __name__ == '__main__':
    colorise.set_color('magenta')

    # All text print after the above call will be in that color...
    print(IPSUM)

    # ...until colorise.set_color is called with no argument or colorise
    # is quit as it will automatically reset colorise



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""ColorFormatter class for colorise.fprint.

This class extends the string.Formatter class.

"""

import re
import string
from colorise.attributes import Attr


class ColorFormatter(string.Formatter):

    """Class for formatting strings containing color syntax.

    As opposed to an ordinary derived string.Formatter, this one does not
    construct and return a formatted string but immediately outputs the
    formatted string. This is necessary to support Windows consoles that cannot
    interpret ANSI escape sequences since they have no way of embedding colors
    into strings but instead set colors through an API call.

    """

    def __init__(self, set_color_func, reset_func):
        """Initialise the color formatter.

        Two OS-dependent functions are passed in for setting and resetting the
        color.

        """
        super().__init__()

        self._autoreset = False
        self._file = None
        self._enabled = True
        self._set_color_func = set_color_func
        self._reset_func = reset_func
        self._attribute_names = Attr.names_with_aliases()

    @property
    def autoreset(self):
        """If True, automatically reset before each color format field."""
        return self._autoreset

    @autoreset.setter
    def autoreset(self, value):
        self._autoreset = value

    @property
    def file(self):
        """The target output handle of the formatter."""
        return self._file

    @file.setter
    def file(self, value):
        self._file = value

    @property
    def enabled(self):
        """Whether colors are enabled or not."""
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        self._enabled = value

    def parse(self, format_string):
        """Parse a format string and generate tokens."""
        # Flush any remaining stuff before resetting colors
        self.file.flush()
        self._reset_func(self.file)

        first_format = True
        tokens = super().parse(format_string)

        for literal_text, field_name, format_spec, conversion in tokens:
            fg, fg_attrs, bg, bg_attrs = self._parse_color_format(field_name)

            if fg or fg_attrs or bg or bg_attrs:
                # Emit any literal text
                yield literal_text, None, None, None

                # Automatically reset colors and attributes if enabled and if
                # it is not the first format we encounter
                if self.autoreset and not first_format:
                    self._reset_func(self.file)

                # Set colors and attributes
                if self.enabled:
                    self._set_color_func(fg, bg, fg_attrs + bg_attrs,
                                         self.file)
            else:
                # Yield tokens as normal
                yield literal_text, field_name, format_spec, conversion

            first_format = False

    def _parse_color_format(self, colors):
        """Parse and extract the color format from a format field."""
        result = [None, [], None, []]

        if not colors:
            return result

        is_fg = False

        for color in colors.split(','):
            color = color.strip()

            if re.match('[fb]g=', color):
                index = 0 if color[0] == 'f' else 2

                if result[index]:
                    raise ValueError('Duplicate {0}ground color format'
                                     .format('back' if index > 0 else 'fore'))

                result[index] = color[3:]
                is_fg = index == 0
            elif color in self._attribute_names:
                # This is an attribute
                result[1 if is_fg else 3].append(Attr.from_name(color))
            else:
                raise ValueError("Unknown color format or attribute '{0}'"
                                 .format(color))

        return result

    def vformat(self, format_string, args, kwargs):
        """Hijack the internals of string.Formatter.vformat.

        Copied (almost) verbatim from the Python 3.7 source code but does not
        return a formatted string.

        """
        used_args = set()
        self._vformat(format_string, args, kwargs, used_args)
        self.check_unused_args(used_args, args, kwargs)

    def _vformat(self, format_string, args, kwargs, used_args,
                 auto_arg_index=0):
        """Hijack the internals of string.Formatter._vformat.

        Copied (almost) verbatim from the Python 3.7 source code but does not
        return a formatted string. The only major changes is that the result
        list normally returned by _vformat is gone and has been replaced with
        direct writes to a stream. Furthermore, nested formats have been
        removed.

        This is extremely brittle if the source ever changes and thus not
        always guaranteed to work. A better work-around is needed.

        """
        tokens = self.parse(format_string)

        for literal_text, field_name, format_spec, conversion in tokens:
            # Output the literal text
            if literal_text:
                self.file.write(literal_text)
                self.file.flush()

            # If there's a field, output it
            if field_name is not None:
                # This is some markup, find the object and do the formatting.
                # Handle arg indexing when empty field_names are given
                if field_name == '':
                    if auto_arg_index is False:
                        raise ValueError('cannot switch from manual field '
                                         'specification to automatic field '
                                         'numbering')
                    field_name = str(auto_arg_index)
                    auto_arg_index += 1
                elif field_name.isdigit():
                    if auto_arg_index:
                        raise ValueError('cannot switch from manual field '
                                         'specification to automatic field '
                                         'numbering')
                    # Disable auto arg incrementing, if it gets used later on,
                    # then an exception will be raised
                    auto_arg_index = False

                # Given the field_name, find the object it references and the
                # argument it came from
                obj, arg_used = self.get_field(field_name, args, kwargs)
                used_args.add(arg_used)

                # Do any conversion on the resulting object
                obj = self.convert_field(obj, conversion)

                # Format the object and append to the result
                self.file.write(self.format_field(obj, format_spec))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Nix color look-up tables (CLUTs) and functions."""

import collections
from colorise.color_tools import closest_color
from colorise.terminal import terminal_name


_COLOR_ESCAPE_CODE = '\x1b['
_COLOR_PREFIX_16 = _COLOR_ESCAPE_CODE + '{0}m'
_COLOR_PREFIX_FG_88 = _COLOR_ESCAPE_CODE + '38;5;{0}m'
_COLOR_PREFIX_BG_88 = _COLOR_ESCAPE_CODE + '48;5;{0}m'
_COLOR_PREFIX_FG_256 = _COLOR_PREFIX_FG_88
_COLOR_PREFIX_BG_256 = _COLOR_PREFIX_BG_88
_COLOR_PREFIX_FG_TRUE_COLOR = _COLOR_ESCAPE_CODE + '38;2;{0}m'
_COLOR_PREFIX_BG_TRUE_COLOR = _COLOR_ESCAPE_CODE + '48;2;{0}m'

_PREFIX_MAP = {
        #       Foreground prefix            Background prefix
        8:     (_COLOR_PREFIX_16,            _COLOR_PREFIX_16),
        16:    (_COLOR_PREFIX_16,            _COLOR_PREFIX_16),
        88:    (_COLOR_PREFIX_FG_88,         _COLOR_PREFIX_BG_88),
        256:   (_COLOR_PREFIX_FG_256,        _COLOR_PREFIX_BG_256),
        2**24: (_COLOR_PREFIX_FG_TRUE_COLOR, _COLOR_PREFIX_BG_TRUE_COLOR)
    }

# System base colors that are assumed to be present for 8 color terminals
_NIX_SYSTEM_COLORS = {
    30: (0x00, 0x00, 0x00),  # black
    31: (0xff, 0x00, 0x00),  # red
    32: (0x00, 0xff, 0x00),  # green
    33: (0xff, 0xff, 0x00),  # yellow
    34: (0x00, 0x00, 0xff),  # blue
    35: (0xff, 0x00, 0xff),  # purple
    36: (0x00, 0xff, 0xff),  # cyan
    37: (0xff, 0xff, 0xff)   # white
}

# System color names as given by colorise
_NIX_SYSTEM_COLOR_NAMES = {
    'black':       30,
    'red':         31,
    'green':       32,
    'yellow':      33,
    'blue':        34,
    'purple':      35,
    'cyan':        36,
    'lightgray':   37,
    'gray':        90,
    'lightred':    91,
    'lightgreen':  92,
    'lightyellow': 93,
    'lightblue':   94,
    'lightpurple': 95,
    'lightcyan':   96,
    'white':       97,
}

# Alias for British english gray and alias for purple
_NIX_SYSTEM_COLOR_NAMES['grey'] = 90
_NIX_SYSTEM_COLOR_NAMES['lightgrey'] = 37
_NIX_SYSTEM_COLOR_NAMES['magenta'] = 35

# xterm 88-color look-up table (based on 88colres.h)
_XTERM_CLUT_88_STEPS = [0x00, 0x8b, 0xcd, 0xff]
_XTERM_CLUT_88_GRAYSCALE = [46, 92, 113, 139, 162, 185, 208, 231]
_XTERM_CLUT_88 = collections.OrderedDict(_NIX_SYSTEM_COLORS)

_XTERM_CLUT_88.update(
    zip(
        range(16, 88),
        [(r, g, b)
         for r in _XTERM_CLUT_88_STEPS
         for g in _XTERM_CLUT_88_STEPS
         for b in _XTERM_CLUT_88_STEPS]
    )
)

_XTERM_CLUT_88.update(
    zip(
        range(80, 89),
        [(g, g, g) for g in _XTERM_CLUT_88_GRAYSCALE])
    )

# xterm 256-color look-up table
_XTERM_CLUT_256_STEPS = [0x00, 0x5f, 0x87, 0xaf, 0xd7, 0xff]
_XTERM_CLUT_256_GRAYSCALE = [8 + 10 * i for i in range(24)]
_XTERM_CLUT_256 = collections.OrderedDict(_NIX_SYSTEM_COLORS)

_XTERM_CLUT_256.update(
    zip(
        range(16, 232),
        [(r, g, b)
         for r in _XTERM_CLUT_256_STEPS
         for g in _XTERM_CLUT_256_STEPS
         for b in _XTERM_CLUT_256_STEPS]
    )
)

_XTERM_CLUT_256.update(
    zip(
        range(232, 256),
        [(g, g, g) for g in _XTERM_CLUT_256_GRAYSCALE]
    )
)


def get_prefix(color_count, bg):
    """Get the color code prefix corresponding to the supported color count."""
    return _PREFIX_MAP[color_count][int(bg)]


def get_clut(color_count):
    """Return the appropriate color look-up table."""
    if terminal_name() == 'iTerm.app' and color_count == 88:
        # Uses the 256 color table for 88 color indices
        return _XTERM_CLUT_256

    return {
        8:   _NIX_SYSTEM_COLORS,
        16:  _NIX_SYSTEM_COLORS,
        88:  _XTERM_CLUT_88,
        256: _XTERM_CLUT_256,
    }[color_count]


def can_redefine_colors():
    """Return whether the terminal allows redefinition of colors."""
    return False


def color_from_name(name, color_count, bg, attributes):
    """Return the color value and color count for a given color name."""
    if name not in _NIX_SYSTEM_COLOR_NAMES:
        raise ValueError("Unknown color name '{0}'".format(name))

    # We just return the 16 color prefix and code. Assumes this can can be done
    # in terminals with more capabilities
    return _COLOR_PREFIX_16, _NIX_SYSTEM_COLOR_NAMES[name] + 10 * int(bg)


def color_from_index(idx, color_count, bg, attributes):
    """Return the color value and color count for a given color index."""
    if idx < 0 or idx > 255:
        raise ValueError('Color index must be in range 0-255 inclusive')

    if color_count > 88:
        # NOTE: Assume we can use 256 color prefixes if we have true-color
        return (_COLOR_PREFIX_BG_256 if bg else _COLOR_PREFIX_FG_256), idx
    elif color_count == 88:
        prefix = _COLOR_PREFIX_BG_88 if bg else _COLOR_PREFIX_FG_88

        if terminal_name() == 'iTerm.app':
            return prefix, idx

        if idx <= 88:
            return prefix, idx
        else:
            # Approximate > 88 color index with 256 color clut
            return prefix,\
                closest_color(_XTERM_CLUT_256[idx], _XTERM_CLUT_88)
    else:
        if idx <= 16:
            return _COLOR_PREFIX_16, idx + 10 * int(bg)
        else:
            # Approximate > 16 color index with 256 color clut
            key = closest_color(_XTERM_CLUT_256[idx], _NIX_SYSTEM_COLORS)

            return _COLOR_PREFIX_16, key + 10 * int(bg)


def get_rgb_color(color_count, bg, rgb, attributes):
    """Get the color for an RGB triple or approximate it if necessary."""
    prefix = get_prefix(color_count, bg)

    if color_count < 2**24:
        # No true-color capabilities and color was given as a true-color value,
        # approximate to closest color given current capabilities
        color_idx = closest_color(rgb, get_clut(color_count))

        if color_count <= 16:
            return prefix, color_idx + 10 * int(bg)
        else:
            return prefix, color_idx

    return prefix, ';'.join(str(c) for c in rgb)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Test the highlight function."""

import colorise
from io import StringIO
import pytest
import os
import sys


def test_highlight():
    text = 'Hello'
    indices = [0, 2, 4]

    colorise.highlight(text, indices, fg='red')
    colorise.highlight(text, indices, fg=201)
    colorise.highlight(text, indices, fg='#a696ff')
    colorise.highlight(text, indices, fg='0xa696ff')
    colorise.highlight(text, indices, fg='hls(0.6919;0.7940;1.0)')
    colorise.highlight(text, indices, fg='hsv(249;41;100)')
    colorise.highlight(text, indices, fg='rgb(167;151;255)')

    colorise.highlight(text, indices, bg='red')
    colorise.highlight(text, indices, bg=201)
    colorise.highlight(text, indices, bg='#a696ff')
    colorise.highlight(text, indices, bg='0xa696ff')
    colorise.highlight(text, indices, bg='hls(0.6919;0.7940;1.0)')
    colorise.highlight(text, indices, bg='hsv(249;41;100)')
    colorise.highlight(text, indices, bg='rgb(167;151;255)')


def test_invalid_highlight():
    text = 'Hello'
    indices = [0, 2, 4]
    colors = [
        'unknown',
        256,
        '#a69ff',
        '0xa69ff',
        'hls(0.6919,0.7940;1.0=',
        'hsv(249;41;-100)',
        'rgb(167;xxx;255)'
    ]

    for color in colors:
        with pytest.raises(ValueError):
            colorise.highlight(text, indices, fg=color)


@pytest.mark.skip_on_windows
def test_highlight_named_output():
    sio = StringIO()
    result = '\x1b[0m\x1b[31mH\x1b[0me\x1b[31ml\x1b[0ml\x1b[31mo\x1b[0m'\
        + os.linesep

    with pytest.redirect_stdout(sio):
        colorise.highlight('Hello', [0, 2, 4], fg='red', file=sys.stdout)
        assert sio.getvalue() == result

    sio = StringIO()
    result = '\x1b[0m\x1b[41mH\x1b[0me\x1b[41ml\x1b[0ml\x1b[41mo\x1b[0m'\
        + os.linesep

    with pytest.redirect_stdout(sio):
        colorise.highlight('Hello', [0, 2, 4], bg='red', file=sys.stdout)
        assert sio.getvalue() == result


@pytest.mark.require_colors(256)
def test_highlight_256_index_output():
    sio = StringIO()
    result = '\x1b[0m\x1b[38;5;201mH\x1b[0me\x1b[38;5;201m'\
             'l\x1b[0ml\x1b[38;5;201mo\x1b[0m' + os.linesep

    with pytest.redirect_stdout(sio):
        colorise.highlight('Hello', [0, 2, 4], fg=201, file=sys.stdout)
        assert sio.getvalue() == result

    sio = StringIO()
    result = '\x1b[0m\x1b[48;5;201mH\x1b[0me\x1b[48;5;201m'\
             'l\x1b[0ml\x1b[48;5;201mo\x1b[0m' + os.linesep

    with pytest.redirect_stdout(sio):
        colorise.highlight('Hello', [0, 2, 4], bg=201, file=sys.stdout)
        assert sio.getvalue() == result


@pytest.mark.require_colors(256**3)
def test_highlight_truecolor_output():
    text = 'Hello'
    indices = [0, 2, 4]
    kwargs = [
        {'fg': '0xa696ff'},
        {'fg': 'hls(0.6919;0.7940;1.0)'},
        {'fg': 'hsv(249;41;100)'},
        {'fg': 'rgb(166;150;255)'},
    ]
    result = '\x1b[0m\x1b[38;2;166;150;255mH\x1b[0me\x1b[38;2;166;150;255m'\
             'l\x1b[0ml\x1b[38;2;166;150;255mo\x1b[0m' + os.linesep

    for kwarg in kwargs:
        sio = StringIO()

        with pytest.redirect_stdout(sio):
            colorise.highlight(text, indices, file=sys.stdout, **kwarg)
            assert sio.getvalue() == result


@pytest.mark.skip_on_windows
def test_highlight_disabled():
    sio = StringIO()

    with pytest.redirect_stdout(sio):
        colorise.highlight('Hello', [0, 2, 4], fg='red', file=sys.stdout,
                           enabled=False)
        assert sio.getvalue() == 'Hello' + os.linesep


@pytest.mark.skip_on_windows
def test_highlight_proper_reset():
    sio = StringIO()
    expected_result = '\x1b[0mH\x1b[44mel\x1b[0ml\x1b[44mo\x1b[0m' + os.linesep

    with pytest.redirect_stdout(sio):
        colorise.set_color(fg='red')
        colorise.highlight('Hello', [1, 2, 4], bg='blue', file=sys.stdout)
        assert sio.getvalue() == expected_result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Demonstration of the various ways of coloring text."""

import random
import colorise


if __name__ == '__main__':
    humpty = ["Humpty Dumpty sat on the wall",
              "Humpty Dumpty had a great fall",
              "All the king's horses and all the king's men",
              "Couldn't put Humpty together again"]

    # Set colors manually
    colorise.set_color('red')
    print(humpty[0])
    colorise.reset_color()

    # Use colorise.cprint
    colorise.cprint(humpty[1], 'green')

    # Use color formatting
    colorise.fprint('{{fg=purple}}{0}'.format(humpty[2]))

    randindices = random.sample(range(len(humpty[3])), 12)

    # You can also highlight different parts of a string...
    colorise.highlight(humpty[3], fg='yellow', bg='red', indices=randindices)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Functions for converting and comparing colors."""

import colorsys
import math
import operator


def hls_to_rgb(hue, lightness, saturation):
    """Convert HLS (hue, lightness, saturation) values to RGB."""
    return tuple(int(math.ceil(c * 255.))
                 for c in colorsys.hls_to_rgb(hue, lightness, saturation))


def hsv_to_rgb(hue, saturation, value):
    """Convert HSV (hue, saturation, value) values to RGB."""
    return tuple(int(c * 255.)
                 for c in colorsys.hsv_to_rgb(hue/360.,
                                              saturation/100.,
                                              value/100.))


def color_difference(rgb1, rgb2):
    """Return the sums of component differences between two colors."""
    return sum(abs(i - j) for i, j in zip(rgb1, rgb2))


def color_distance(rgb1, rgb2):
    """Compute the Euclidian distance between two colors."""
    r1, g1, b1 = rgb1
    r2, g2, b2 = rgb2

    return math.sqrt((r2 - r1)**2 + (g2 - g1)**2 + (b2 - b1)**2)


def closest_color(rgb, clut):
    """Return the CLUT index of the closest RGB color to a given RGB tuple."""
    # Generate a list of tuples of CLUT indices and the color difference value
    indexed_diffs = ((idx, color_difference(rgb, clut[idx])) for idx in clut)

    return min(indexed_diffs, key=operator.itemgetter(1))[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Class for Windows handles."""


class WinHandle:
    """Represents a Windows stream handle."""

    def __init__(self, handle):
        """Initialise the Windows handle."""
        self._handle = handle
        self._console_mode = 0
        self._fg = -1
        self._bg = -1
        self._default_fg = -1
        self._default_bg = -1

    @property
    def handle(self):
        """Return the internal Windows handle."""
        return self._handle

    @property
    def fg(self):
        """Return the current foreground color set for the handle."""
        return self._fg

    @fg.setter
    def fg(self, value):
        self._fg = value

    @property
    def bg(self):
        """Return the current background color set for the handle."""
        return self._bg

    @bg.setter
    def bg(self, value):
        self._bg = value

    @property
    def default_fg(self):
        """Return the default foreground color set for the handle."""
        return self._default_fg

    @default_fg.setter
    def default_fg(self, value):
        self._default_fg = value

    @property
    def default_bg(self):
        """Return the default background color set for the handle."""
        return self._default_bg

    @default_bg.setter
    def default_bg(self, value):
        self._default_bg = value

    @property
    def console_mode(self):
        """Return the current console mode for the handle."""
        return self._console_mode

    @console_mode.setter
    def console_mode(self, value):
        self._console_mode = value

    def __str__(self):
        return "{0}({1}, {2})".format(self.__class__.__name__,
                                      self.fg, self.bg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Show how to use different colorspaces."""

import colorise


if __name__ == "__main__":
    colorise.cprint('Using color names', fg='red')
    colorise.cprint('Using color indices', fg=196)
    colorise.cprint("Using hex codes with '#'", fg='#27ee45')
    colorise.cprint("Using hex codes with '0x'", fg='0x27ee45')
    colorise.cprint('Using HLS', fg='hls(0.3; 0.2; 0.1)')
    colorise.cprint("Using HSV", fg='hsv(65; 23;90)')
    colorise.cprint('Using RGB', fg='rgb(57;95;200)')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Attributes supported by colorise.

Depending on your terminal's capabilities, some of these attributes may have no
effect.

"""

from enum import Enum


class Attr(Enum):
    """Console attributes.

    Each attribute's value is defined via its ANSI escape code.

    """

    Reset = 0
    Bold = 1
    Intense = 1  # Alias for Bold from Windows terminology
    Faint = 2
    Italic = 3
    Underline = 4
    Blink = 5
    Reverse = 7

    @classmethod
    def from_name(cls, attribute):
        """Return an attribute object from its name.

        E.g. 'bold' -> Attr.Bold

        """
        return cls[attribute.title()]

    @classmethod
    def names(cls):
        """Return a set of all attribute names."""
        return set(attr.name.lower() for attr in cls)

    @classmethod
    def names_with_aliases(cls):
        """Return a set of all attribute names with aliases."""
        return set(attr.lower() for attr in cls.__members__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Main color look-up table (CLUT) functions.

All look-up tables presented here are entirely based on research and the
author's own knowledge as it is impossible to ensure complete cross-platform
support across all kinds of terminals. Therefore, there is no guarantee that
any of these tables will be accurate.

If you have corrections or suggestions, please submit an issue.

"""

import re
from colorise.color_tools import hls_to_rgb, hsv_to_rgb

_DELIMITER = ';'
_RGB_RE = re.compile(r'^(rgb)?\((\d{1,3}' +
                     _DELIMITER +
                     r'\s*\d{1,3}' +
                     _DELIMITER +
                     r'\s*\d{1,3})\)$')
_HEX_RE = re.compile(r'^(0x|#)?(([0-9a-fA-F]{2}){3})$')
_HSV_RE = re.compile(r'^(hsv)\((\d+' +
                     _DELIMITER +
                     r'\s*\d+' +
                     _DELIMITER +
                     r'\s*\d+)\)$')
_HLS_RE = re.compile(r'^(hls)\((\d+(\.\d+)?' +
                     _DELIMITER +
                     r'\s*\d+(\.\d+)?' +
                     _DELIMITER +
                     r'\s*\d+(\.\d+)?)\)$')

# Supported formats of colorise. The order matters!
_FORMATS = [
    (_RGB_RE.match, 'rgb'),
    (str.isdigit,   'index'),
    (str.isalpha,   'name'),
    (_HEX_RE.match, 'hex'),
    (_HLS_RE.match, 'hls'),
    (_HSV_RE.match, 'hsv')
]


def match_color_formats(value):
    """Return the color format of the first format to match the given value."""
    for matcher, name in _FORMATS:
        match = matcher(str(value))

        if match:
            return match, name

    return None, None


def get_color(value, color_count, cluts, bg=False, attributes=[]):
    """Return the color given by a color format."""
    match, colorspace = match_color_formats(value)

    if colorspace == 'name':
        # Color was given as text, e.g. 'red'
        return cluts.color_from_name(value, color_count, bg, attributes)
    elif colorspace == 'index':
        # Color is a 8, 16, 88 or 256 color index
        return cluts.color_from_index(int(value), color_count, bg, attributes)
    elif colorspace == 'hex':
        value = match.group(2)
        rgb = [int(value[i:i+2], 16) for i in range(0, 6, 2)]
    elif colorspace == 'hsv':
        rgb = hsv_to_rgb(*[float(c) for c in match.group(2).split(_DELIMITER)])
    elif colorspace == 'hls':
        rgb = hls_to_rgb(*[float(c) for c in match.group(2).split(_DELIMITER)])
    elif colorspace == 'rgb':
        rgb = [int(c.strip()) for c in match.group(2).split(_DELIMITER)]
    else:
        raise ValueError("Unknown color format '{0}'".format(value))

    return cluts.get_rgb_color(color_count, bg, rgb, attributes)



"""colorise module setup script for distribution."""

from setuptools import setup
import os


def get_version(filename):
    with open(filename) as fh:
        for line in fh:
            if line.startswith('__version__'):
                return line.split('=')[-1].strip()[1:-1]


setup(
    name='colorise',
    version=get_version(os.path.join('colorise', '__init__.py')),
    author='Alexander Asp Bock',
    author_email='albo.developer@gmail.com',
    platforms='Platform independent',
    python_requires='>=3.4',
    description=('Easily print colored text to the console'),
    license='BSD 3-Clause License',
    keywords='text, color, colorise, colorize, console, terminal',
    packages=['colorise', 'colorise.win', 'colorise.nix'],
    package_data={'colorise': ['tests', 'examples']},
    url='https://github.com/MisanthropicBit/colorise',
    project_urls={
        'Issue Tracker': 'https://github.com/MisanthropicBit/colorise/issues',
        'Documentation': 'https://colorise.readthedocs.io/en/latest/'
    },
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Utilities',
        'Topic :: Terminals',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: PyPy'
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Test redefinition of colors."""

import colorise
import pytest


@pytest.mark.skip_on_windows
def test_redefine_colors_error():
    with pytest.raises(colorise.error.NotSupportedError):
        colorise.redefine_colors({})



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""A simple function that illustrates how to use colorise.highlight."""

import colorise
from colorise.attributes import Attr


def cprint_missing(ls1, ls2, color, attributes):
    """Print a padding placeholder for strings of different lengths."""
    if ls2 > ls1:
        colorise.cprint('_' * (ls2-ls1), fg=color, attributes=attributes)
    else:
        print()


def colordiff(s1, s2, color):
    """Highlight the characters in s2 that differ from those in s1."""
    ls1, ls2 = len(s1), len(s2)
    diff_indices = [i for i, (a, b) in enumerate(zip(s1, s2)) if a != b]

    print(s1, end='')
    cprint_missing(ls1, ls2, color, [Attr.Bold])

    colorise.highlight(s2, indices=diff_indices, attributes=[Attr.Bold],
                       fg=color, end='')

    cprint_missing(ls1, ls2, color, [Attr.Bold])


if __name__ == '__main__':
    colordiff('Highlight the differences',
              'Hightight tke sifferenc',
              'red')

    print()

    colordiff('I wonder if these strings are any different?',
              'A wonder iF thzse xxxings are xxy different???',
              'red')
